@@ -11,25 +11,21 @@
 ## "IrRep" code and under terms of GNU General Public license v3 #
 ## see LICENSE file in the                                       #
 ##                                                               #
-##  Written by Stepan Tsirkin, University of Zurich.             #
-##  e-mail: stepan.tsirkin@physik.uzh.ch                         #
+##  Written by Stepan Tsirkin                                    #
+##  e-mail: stepan.tsirkin@ehu.eus                               #
 ##################################################################
 
 """
 Defines the command line interface to "irrep".
 """
 
-import sys
 import numpy as np
-import datetime
-import math
 
 import click
 from monty.serialization import dumpfn, loadfn
 
-from .spacegroup import SpaceGroup
 from .bandstructure import BandStructure
-from .utility import str2bool, str2list, short
+from .utility import sort_vectors, str2list, short, log_message
 from . import __version__ as version
 
 
@@ -99,6 +95,13 @@ do not hesitate to contact the author:
     "A value of 50 eV is recommended. If not set, will default to "
     "the cut-off used in the DFT calculation.",
 )
+@click.option(
+    "-correct_Ecut0",
+    type=float,
+    default=0.,
+    help="In case of VASP, if you get an error like ' computed ncnt=*** != input nplane=*** ', "
+        "try to set this parameter to a small positive or negative value (usually of order  +- 1e-7)"
+)
 @click.option(
     "-fWAV",
     type=str,
@@ -119,11 +122,23 @@ do not hesitate to contact the author:
     help="Filename for wavefunction in ABINIT WFK format. "
     'Only used if code is "abinit".',
 )
+@click.option(
+    "-gpaw_calc",
+    type=str,
+    help="Filename for gpaw calculator. "
+    'Only used if code is "gpaw".',
+)
 @click.option(
     "-prefix",
     type=str,
     help="Prefix used for Quantum Espresso calculations (data should be in prefix.save) or seedname of Wannier90 files. ",
 )
+@click.option(
+    "-from_sym_file",
+    type=str,
+    help="if present, the symmetry operations will be read from this file, instead of those computed by spglib",
+)
+
 @click.option(
     "-IBstart",
     type=int,
@@ -140,7 +155,7 @@ do not hesitate to contact the author:
 )
 @click.option(
     "-code",
-    type=click.Choice(["vasp", "abinit", "espresso", "wannier90"]),
+    type=click.Choice(["vasp", "abinit", "espresso", "wannier90", "gpaw"]),
     default="vasp",
     help="Set which electronic structure code to interface with. If using ABINIT, always use "
     '"istwfk=1".',
@@ -180,8 +195,9 @@ do not hesitate to contact the author:
 )
 @click.option(
     "-isymsep",
-    help="Index of the symmetry to separate the eigenstates. Works well only for norm-conserving "
-    "potentials as in ABINIT.",
+    help="Index of the symmetry to separate the eigenstates. "
+    "with new method works with any code/pseudopotential"
+    "Previously worked well only for norm-conserving potentials.",
 )
 @click.option(
     "-onlysym",
@@ -189,6 +205,19 @@ do not hesitate to contact the author:
     default=False,
     help="Only calculate the symmetry operations",
 )
+@click.option(
+    "-writesym",
+    flag_value=True,
+    default=False,
+    help="write the prefix.sym file needed for the Wannier90 sitesym calculations",
+)
+@click.option(
+    "-alat",
+    type=float,
+    default=None,
+    help="for writesym - the alat parameter. For QE, it is read from the prefix.save/data-file-schema.xml"
+    "For other codes needs to be provided. (To be honest, the .sym file is useless for other codes for now, but still ..)",
+)
 @click.option("-ZAK", flag_value=True, default=False, help="Calculate Zak phase")
 @click.option(
     "-WCC", flag_value=True, default=False, help="Calculate Wannier charge centres"
@@ -199,12 +228,6 @@ do not hesitate to contact the author:
     default=False,
     help="Write gnuplottable files with all symmetry eigenvalues",
 )
-@click.option(
-    "-plotFile", 
-    type=str, 
-    help="file where bands for plotting will be written."
-    "In development...!"
-)
 @click.option("-EF", 
     type=str, 
     default='0.0',
@@ -222,8 +245,8 @@ do not hesitate to contact the author:
 @click.option(
     "-groupKramers", 
     flag_value=True, 
-    default=True, 
-    help="Group wave-functions in pairs of Kramers. Default: True."
+    default=False, 
+    help="Group wave-functions in pairs of Kramers. Default: False."
 )
 @click.option(
     "-symmetries", 
@@ -247,12 +270,36 @@ do not hesitate to contact the author:
           "to the conventional cell."
           "Default: False, unless -kpnames was specified in CLI"
           )
+    )
+@click.option("-trans_thresh",
+    type=float,
+    default=1e-5,
+    help=("Threshold to compare translational parts of symmetries."
+          "Default: 1e-5"
+          )
+)
+@click.option("-v",
+              count=True,
+              default=1,
+              help=("Verbosity flag. -vv: very detailed info, recommended "
+                    "when you get an error. -v (default for CLI): info about "
+                    "some decissions taken internaly through the code's "
+                    "execution, recommended when the code runs without "
+                    "errors, but the result is not what you expected. If you "
+                    "don't set this tag, you will get the basic info.")
+)
+@click.option("-json_file",
+                 type=str,
+                    default="irrep-output.json",
+                    help="File to save the output in JSON format. (without "
+                    "extension, the '.json' will be added automatically)"
 )
 def cli(
     ecut,
     fwav,
     fpos,
     fwfk,
+    gpaw_calc,
     prefix,
     ibstart,
     ibend,
@@ -264,17 +311,23 @@ def cli(
     shiftuc,
     isymsep,
     onlysym,
+    writesym,
+    alat,
+    from_sym_file,
     zak,
     wcc,
     plotbands,
-    plotfile,
     ef,
     degenthresh,
     groupkramers,
     symmetries,
     suffix,
     config,
-    searchcell
+    searchcell,
+    correct_ecut0,
+    trans_thresh,
+    v,
+    json_file
 ):
     """
     Defines the "irrep" command-line tool interface.
@@ -286,31 +339,26 @@ def cli(
         refuc = np.array(refuc.split(","), dtype=float).reshape((3, 3))
     if shiftuc:
         shiftuc = np.array(shiftuc.split(","), dtype=float).reshape(3)
-    
+
+    # rename the v flag to verbosity, to avoid overlap with local variables
+    verbosity = v    
     # Warning about kpnames
     if kpnames is not None:
         searchcell = True
+
     elif not searchcell:
-        print(("Warning: transformation to the convenctional unit "
+        msg = ("Warning: transformation to the convenctional unit "
                "cell will not be calculated, nor its validity checked "
                "(if given). See the description of flag -searchcell "
                "on:\n"
                "irrep --help"
-               ))
-        print(("Warning: -kpnames not specified. Only traces of "
+               )
+        log_message(msg, verbosity, 1)
+        msg = ("Warning: -kpnames not specified. Only traces of "
                "symmetry operations will be calculated. Remember that "
                "-kpnames must be specified to identify irreps"
-               ))
-
-    #if kpnames is None:
-    #    identify_irreps = False
-    #    print(("Warning: kpnames not specified. Only traces of "
-    #           "symmetry operations will be calculated. Remember that "
-    #           "kpnames must be specified to identify irreps"
-    #           )
-    #          )
-    #else:
-    #    identify_irreps = True
+               )
+        log_message(msg, verbosity, 1)
 
     # parse input arguments into lists if supplied
     if symmetries:
@@ -322,17 +370,15 @@ def cli(
     if kpnames:
         kpnames = kpnames.split(",")
 
-    try:
-        print(fwfk.split("/")[0].split("-"))
-        preline = " ".join(s.split("_")[1] for s in fwfk.split("/")[0].split("-")[:3])
-    except Exception as err:
-        print(err)
-        preline = ""
-    json_data = {}
+    if isymsep or wcc or zak:
+        save_wf = True
+    else:
+        save_wf = False
 
     bandstr = BandStructure(
         fWAV=fwav,
         fWFK=fwfk,
+        calculator_gpaw=gpaw_calc,
         prefix=prefix,
         fPOS=fpos,
         Ecut=ecut,
@@ -340,15 +386,23 @@ def cli(
         IBend=ibend,
         kplist=kpoints,
         spinor=spinor,
+        calculate_traces=True,
         code=code,
         EF=ef,
         onlysym=onlysym,
         refUC = refuc,
         shiftUC = shiftuc,
         search_cell = searchcell,
+        degen_thresh=degenthresh,
+        save_wf=save_wf,
+        verbosity=verbosity,
+        from_sym_file=from_sym_file
     )
 
-    json_data ["spacegroup"] = bandstr.spacegroup.show(symmetries=symmetries)
+    bandstr.spacegroup.show()
+
+    if writesym:
+        bandstr.spacegroup.write_sym_file(filename=prefix+".sym", alat=alat)
 
     if onlysym:
         exit()
@@ -358,30 +412,63 @@ def cli(
                 bandstr.spacegroup.str()
                 )
 
+    # Identify irreps. If kpnames wasn't set, all will be labelled as None
+    bandstr.identify_irreps(kpnames, verbosity=verbosity)
+
+    # Temporary, until we make it valid for isymsep
+    bandstr.write_characters()
+
+    # Write irreps.dat file
+    if kpnames is not None:
+        bandstr.write_irrepsfile()
+
+    # Write trace.txt file
+    bandstr.write_trace()
+
+    # Temporary, until we make it valid for isymsep
+    json_data = {}
+    json_data ["spacegroup"] = bandstr.spacegroup.json(symmetries=symmetries)
+    json_bandstr = bandstr.json()
+    json_data['characters and irreps'] = [{"subspace": json_bandstr}]
+
+    # Separate in terms of symmetry eigenvalues
     subbands = {(): bandstr}
 
     if isymsep is not None:
         json_data["separated by symmetry"]=True
         json_data["separating symmetries"]=isymsep
+        tmp_subbands = {}
         for isym in isymsep:
-            print("Separating by symmetry operation # ", isym)
-            subbands = {
-                tuple(list(s_old) + [s_new]): sub
-                for s_old, bands in subbands.items()
-                for s_new, sub in bands.Separate(
-                    isym, degen_thresh=degenthresh, groupKramers=groupkramers
-                ).items()
-            }
+            print("\n-------- SEPARATING BY SYMMETRY # {} --------".format(isym))
+            for s_old, bs in subbands.items():
+                separated = bs.Separate(isym, groupKramers=groupkramers, verbosity=verbosity)
+                for s_new, bs_separated in separated.items():
+                    tmp_subbands[tuple(list(s_old) + [s_new])] = bs_separated
+            subbands = tmp_subbands
+        json_data["characters and irreps"]=[]
+        
+        # sort to have consistency between runs
+        for k in sort_vectors(subbands.keys()):
+            sub = subbands[k]
+            if isymsep is not None:
+                print(
+                    "\n\n\n\n ################################################ \n\n\n NEXT SUBSPACE:  ",
+                    " , ".join(
+                        "sym # {0} -> eigenvalue {1}".format(s, short(ev)) for s, ev in zip(isymsep, k)
+                    ),
+                )
+                sub.write_characters()
+                json_data["characters and irreps"].append({"symmetry eigenvalues": np.array(k) , "subspace": sub.json(symmetries)})
     else :
         json_data["separated by symmetry"]=False
         
 
+    dumpfn(json_data, json_file, indent=4)
+
     if zak:
         for k in subbands:
             print("symmetry eigenvalue : {0} \n Traces are : ".format(k))
-            subbands[k].write_characters(
-                degen_thresh=0.001, symmetries=symmetries
-            )
+            subbands[k].write_characters()
             print("symmetry eigenvalue : {0} \n Zak phases are : ".format(k))
             zak = subbands[k].zakphase()
             for n, (z, gw, gc, lgw) in enumerate(zip(*zak)):
@@ -404,33 +491,8 @@ def cli(
             )
         exit()
 
-    bandstr.write_trace(
-        degen_thresh=degenthresh,
-        symmetries=symmetries,
-    )
-    json_data["characters_and_irreps"]=[]
-    for k, sub in subbands.items():
-        if isymsep is not None:
-            print(
-                "\n\n\n\n ################################################ \n\n\n next subspace:  ",
-                " , ".join(
-                    "{0}:{1}".format(s, short(ev)) for s, ev in zip(isymsep, k)
-                ),
-            )
-        plotfile=None # being implemented, not finished yet...
-        characters = sub.write_characters(
-            degen_thresh=degenthresh,
-            symmetries=symmetries,
-            kpnames=kpnames,
-            preline=preline,
-            plotFile=plotfile,
-        )
-        json_data["characters_and_irreps"].append( {"symmetry_eigenvalues":k , "subspace": characters  }  )
-#        json_data["characters_and_irreps"].append( [k ,characters  ]  )
-
     if plotbands:
-        json_data["characters_and_irreps"] = {}
-        print("plotbands = True --> writing bands")
+        print("\nplotbands = True --> writing bands")
         for k, sub in subbands.items():
             if isymsep is not None:
                 print(
@@ -448,20 +510,6 @@ def cli(
                     )
                     + ".dat"
                 )
-                fname1 = (
-                    "bands-sym-"
-                    + suffix
-                    + "-"
-                    + "-".join(
-                        "{0}:{1}".format(s, short(ev)) for s, ev in zip(isymsep, k)
-                    )
-                    + ".dat"
-                )
             else:
                 fname = "bands-{0}.dat".format(suffix)
-                fname1 = "bands-sym-{0}.dat".format(suffix)
-            with open(fname, "w") as f:
-                f.write(sub.write_bands())
-            sub.write_trace_all(degenthresh, fname=fname1)
-
-    dumpfn(json_data,"irrep-output.json",indent=4)
+            sub.write_plotfile(fname)
